#include <WiFi.h>
#include <LiquidCrystal_I2C.h>
#include <DHT.h>
#include <ArduinoJson.h>
#include <freertos/FreeRTOS.h>
#include <freertos/task.h>
#include <Arduino.h>
#include <ESPAsyncWebServer.h>

#define DHT_PIN 15
#define DHT_TYPE DHT22

#define LCD_ADDRESS 0x27
#define LCD_CHARS 16
#define LCD_LINES 2

#define JSON_SIZE 1536

#define NETWORK ""
#define PASSWORD ""

char WEBSITE[] = "api.weatherapi.com";
char PATH[] = ""; // here add path with api key
int HTTP_PORT = 80;
String REQ_METHOD = "GET";

LiquidCrystal_I2C lcd(LCD_ADDRESS, LCD_CHARS, LCD_LINES);
DHT dht(DHT_PIN, DHT_TYPE);
AsyncWebServer server(HTTP_PORT);

float tempInside = 0, humInside = 0;
float tempOutside = 0, humOutside = 0, windOutside = 0;
const char* weather;

bool firstCall = true;
unsigned long previousMillis = 0;
const long interval = 1800000;

DynamicJsonDocument doc(JSON_SIZE);

void connectToWiFi() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(NETWORK, PASSWORD);

  while (WiFi.status() != WL_CONNECTED) {
    Serial.print('.');
    delay(100);
  }

  Serial.println();
  Serial.println(WiFi.localIP());

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Connected");
  lcd.setCursor(0, 1);
  lcd.print("To Internet..");

  Serial.println("Connected to internet!");
  delay(2000);
}

void displayLCDInside() {
  lcd.clear();

  lcd.setCursor(0, 0);
  lcd.print("Temp: ");
  lcd.printf("%.1f", tempInside);
  lcd.write(0xDF);
  lcd.print("C");

  lcd.setCursor(0, 1);
  lcd.print("Humidity: ");
  lcd.printf("%.1f", humInside);
  lcd.print(" %");
}

void displayLCDOutside() {
  lcd.clear();

  lcd.setCursor(0, 0);
  lcd.print(weather);

  lcd.setCursor(0, 1);
  lcd.print("Temp: ");
  lcd.printf("%.1f", tempOutside);
  lcd.write(0xDF);
  lcd.print("C");

  delay(10000);

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Humidity: ");
  lcd.printf("%.1f", humOutside);
  lcd.print(" %");

  lcd.setCursor(0, 1);
  lcd.print("Wind: ");
  lcd.print(windOutside);
  lcd.print(" km/h");
}

void apiCall() {
  WiFiClient client;

  if (client.connect(WEBSITE, HTTP_PORT)) {
    Serial.println("Connected to " + String(WEBSITE));
    client.println(REQ_METHOD + " " + String(PATH) + " HTTP/1.1");
    client.println("Host: " + String(WEBSITE));
    client.println(F("Cache-Control: no-cache"));
    client.println("Connection: close");
    client.println();

    // wait for response
    delay(1000);

    char status[32] = { 0 };

    client.readBytesUntil('\r', status, sizeof(status));

    Serial.println("Status:" + String(status));

    // if response status code is ok
    if (strcmp(status, "HTTP/1.1 200 OK") != 0) {
      Serial.println("Didn't receive status code 200!");
    } else {
      Serial.println("OK!");
    }

    // skip the headers
    char endOfHeaders[] = "\r\n\r\n";
    if (!client.find(endOfHeaders)) {
      Serial.println(F("Invalid response"));
      return;
    }

    // skip random chars until json starts
    while (client.available() && client.peek() != '{') {
      char c = 0;
      client.readBytes(&c, 1);
    }

    // error deserializing
    DeserializationError error = deserializeJson(doc, client);

    if (!error) {
      JsonObject current = doc["current"];

      // get fields from json
      tempOutside = current["temp_c"];
      windOutside = current["wind_kph"];
      humOutside = current["humidity"];

      JsonObject current_condition = current["condition"];

      weather = current_condition["text"];
    } else {

      Serial.print("Deserialization failed: ");
      Serial.println(error.f_str());
    }

    client.stop();

  } else {
    Serial.println("Connection to " + String(WEBSITE) + " failed");
  }

  delay(5000);
}

void displayLCDTask(void* parameter) {
  while (true) {
    tempInside = dht.readTemperature();
    humInside = dht.readHumidity();

    displayLCDInside();
    vTaskDelay(pdMS_TO_TICKS(10000));
    displayLCDOutside();
    vTaskDelay(pdMS_TO_TICKS(10000));
  }
}

void apiCallTask(void* parameter) {
  while (true) {
    unsigned int currentMillis = millis();

    if (firstCall) {
      apiCall();
      firstCall = false;
      previousMillis = currentMillis;
    }

    // Check if 30 minutes have passed since the last call to apiCall()
    if (currentMillis - previousMillis >= interval) {
      apiCall();
      previousMillis = currentMillis;
    }

    vTaskDelay(pdMS_TO_TICKS(60000));
  }
}

void setup() {

  Serial.begin(115200);

  lcd.init();
  lcd.backlight();

  lcd.setCursor(0, 0);
  lcd.print("Init LCD...");

  delay(2000);

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Done!");

  dht.begin();

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Init DHT...");

  delay(2000);

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Done!");

  delay(2000);

  connectToWiFi();

  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Starting");

  lcd.setCursor(0, 1);
  lcd.print("Server...");

  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    String color;

    if (!strncmp(weather, "Sunny", 5)) {
      color = "gold";
    } else if (!strncmp(weather, "Clear", 5)) {
      color = "lightblue";
    } else {
      color = "grey";
    }

    String html = "<!DOCTYPE HTML>"
                  "<html>"
                  "<head>"
                  "<meta charset=\"UTF-8\">"
                  "<meta name=\"viewport\" content=\"width=device-width, initial-scale=1.0\">"
                  "<title>My ESP32 Web Server</title>"
                  "</head>"
                  "<body style=\"display:flex; justify-content:center; flex-direction:column; align-items:center; width:100%; text-align:center;\">"
                  "<h1>Weather Station Web Server</h1>"
                  "<h2>Temperature and Humidity at Home:</h2>"
                  "<h3 style=\"color:red;\">Temperature: " + String(tempInside) + "&deg; C</h3>"
                  "<h3 style=\"color:blue;\">Humidity: " + String(humInside) + "%</h3><br>"
                  "<div><h2 style=\"color:black; display:inline;\">Weather in Bucharest:  </h2>"
                  "<h2 style=\"color:" + color + "; display:inline;\">" + String(weather) + "</h2></div>"
                  "<h3 style=\"color:red;\">Temperature: " + String(tempOutside) + "&deg; C</h3>"
                  "<h3 style=\"color:blue;\">Humidity: " + String(humOutside) + "%</h3>"
                  "<h3 style=\"color:green;\">Wind Speed: " + String(windOutside) + " km/h</h3>"
                  "</body>"
                  "</html>";
    request->send(200, "text/html", html);
  });

  // Start server
  server.begin();

  delay(2000);

  // create tasks in rtos
  xTaskCreate(apiCallTask, "ApiCallTask", 10000, NULL, 3, NULL);
  xTaskCreate(displayLCDTask, "DisplayLCDTask", 10000, NULL, 2, NULL);
}

void loop() {
  
}
